"""
App settings management.
"""

from typing import Any, Dict, Optional
from PyQt5 import QtCore


class SettingProxy(QtCore.QObject):
    
    valueChanged = QtCore.pyqtSignal(object)
    
    def __init__(self, settings: QtCore.QSettings, key: str, type=str, default=None):
        super().__init__()
        
        self._settings = settings
        self._key = key
        self._type = type
        self._default = default
        
        if default and not self.value:
            self.value = default
    
    @property
    def value(self) -> Any:
        return self._settings.value(self._key, type=self._type, defaultValue=self._default)
    
    @value.setter
    def value(self, value: Any):
        self._settings.setValue(self._key, value)
        self.valueChanged.emit(value)


class SettingsManager:
    """
    Settings manager. Maintains a namespace of settings proxies.
    
    Example of a str-valued setting with no default:
    
    >>> settings = SettingsManager.get_instance()
    >>> settings.some_str = 'mysection/mykey'
    >>> settings.some_str.value = 'asdf'  # settings.some_str emits valueChanged with 'asdf'
    >>> settings.some_str.value  # returns 'asdf'
    
    Example of an int-valued setting with a default:
    >>> settings.some_int = ('mysection/mykey2', int, 42)
    >>> settings.some_int.value  # returns 42
    >>> settings.some_int.value = 24  # settings.some_int emits valueChanged with 24
    """
    
    _instance: 'SettingsManager' = None
    
    @classmethod
    def get_instance(cls) -> 'SettingsManager':
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    
    def __init__(self, settings: Optional[QtCore.QSettings] = None):
        self._settings = settings or QtCore.QSettings()
        self._proxies: Dict[str, SettingProxy] = {}
    
    def __getattribute__(self, __name: str) -> Any:
        if __name in ('_settings', '_proxies'):
            return super().__getattribute__(__name)
        elif __name in self._proxies:
            return self._proxies[__name]
        else:
            raise AttributeError(f'No such setting: {__name}')
    
    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name in ('_settings', '_proxies'):
            super().__setattr__(__name, __value)
        elif isinstance(__value, SettingProxy):
            self._proxies[__name] = __value
        elif isinstance(__value, str):
            self._proxies[__name] = SettingProxy(self._settings, __value)
        elif isinstance(__value, tuple):
            self._proxies[__name] = SettingProxy(self._settings, __value[0], *__value[1:])



from PyQt5 import QtWidgets, QtCore

from boxjelly.ui.track.Timeline import Timeline
from boxjelly.ui.track.TrackHeadersView import TrackHeadersView
from boxjelly.ui.track.TrackGraphicsView import TrackGraphicsView
from boxjelly.models.TrackListModel import TrackListModel


class TrackPanel(QtWidgets.QWidget):
    """Track panel widget"""
    
    frameSelected = QtCore.pyqtSignal(int)
    
    def __init__(self, parent, track_model: TrackListModel, selection_model: QtCore.QItemSelectionModel):
        super().__init__(parent)
        
        self._track_model = track_model
        self._selection_model = selection_model
        
        # Set minimum size
        self.setMinimumSize(800, 0)
        
        # Set default size
        self.setGeometry(0, 0, 800, 100)
        
        # Create timeline
        self.timeline = Timeline(self, 300)
        
        # Create track view
        self.track_view = TrackGraphicsView(self, self._track_model, self._selection_model)
        
        # Create track headers view
        self.track_headers_view = TrackHeadersView(self)
        self.track_headers_view.setModel(self._track_model)
        self.track_headers_view.setSelectionModel(self._selection_model)
        
        # Connect signals and slots
        self.track_view.frameScrolled.connect(self.timeline.set_frame)
        self.track_view.verticalScrollBarChanged.connect(self.track_headers_view.set_vertical_scroll)
        self.track_view.scaleChanged.connect(lambda x_scale, y_scale: self.timeline.set_frame_scale(x_scale))
        self.timeline.frameSelected.connect(self.frameSelected.emit)
        
        self.arrange()
    
    def arrange(self):
        """Arrange the track panel"""
        # Set vertical layout
        self.setLayout(QtWidgets.QGridLayout())
        
        # Arrange timeline
        self.layout().addWidget(self.timeline, 0, 1, 1, 1)
        
        # Arrange track headers list view
        self.layout().addWidget(self.track_headers_view, 1, 0, 1, 1)
        
        # Arrange track view
        self.layout().addWidget(self.track_view, 1, 1, 1, 1)
        
        self.layout().setRowStretch(0, 0)
        self.layout().setRowStretch(1, 1)
        self.layout().setColumnStretch(0, 0)
        self.layout().setColumnStretch(1, 1)
    
    def show_frame(self, frame: int):
        """Show data corresponding to a particular frame number"""
        self.track_view.show_frame(frame)



from PyQt5 import QtWidgets


class WidgetComboBox(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        
        # Create combo box
        self.combo_box = QtWidgets.QComboBox()
        
        # Create placeholder widget
        self.widget = QtWidgets.QWidget()
        
        # Connect combo box change signal
        self.combo_box.activated.connect(self.on_combo_box_changed)
        
        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Add combo box to layout
        self.layout().addWidget(self.combo_box)
    
    def on_combo_box_changed(self, index):
        """Slot for combo box change signal"""
        # Get the widget
        widget = self.combo_box.itemData(index)
        
        # Set the widget
        self.set_widget(widget)
    
    def set_widget(self, widget):
        """Set the widget"""
        # Hide the current widget
        self.widget.hide()
        
        # Set the widget
        self.widget = widget
        
        # Show the widget
        self.widget.show()
    
    def add_widget(self, widget, name):
        """Add a widget to the combo box"""
        # Add the widget to the combo box
        self.combo_box.addItem(name, widget)
        self.layout().addWidget(widget)
        widget.hide()
    
    def refresh(self):
        """Refresh the loaded widget display status"""
        self.on_combo_box_changed(self.combo_box.currentIndex())


from PyQt5 import QtWidgets, QtGui

from boxjelly.ui.MediaSelectionBox import MediaSelectionBox
from boxjelly.ui.TrackSelectionBox import TrackSelectionBox


class OpenDialog(QtWidgets.QDialog):
    """Dialog to get a video and track source."""
    
    def __init__(self, parent):
        super().__init__(parent)
        
        # Style
        self.setWindowTitle('Open')
        self.setWindowIcon(QtGui.QIcon(':/icons/boxjelly_logo.svg'))
        self.setModal(True)
        self.resize(500, 0)
        
        # Create the selection boxes
        self._media_selection_box = MediaSelectionBox()
        self._track_selection_box = TrackSelectionBox()
        
        # Create the dialog buttons
        self._button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)
        
        # Arrange the widgets
        self._arrange()
        
    def _arrange(self):
        self.setLayout(QtWidgets.QVBoxLayout())
        
        self.layout().addWidget(self._media_selection_box)
        self.layout().addWidget(self._track_selection_box)
        self.layout().addWidget(self._button_box)
    
    @property
    def media_content(self):
        return self._media_selection_box.media_content
    
    @property
    def track_io(self):
        return self._track_selection_box.io



from PyQt5 import QtWidgets, QtCore, QtGui

from boxjelly.models.TrackListModel import TrackListModel


class TrackHeadersDelegate(QtWidgets.QStyledItemDelegate):
    
    def paint(self, painter: QtGui.QPainter, opt: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex):
        self.initStyleOption(opt, index)
        
        rect = opt.rect
        
        painter.save()
        
        # Get the track ID and label
        id = index.data(TrackListModel.IDRole)
        label = index.data(TrackListModel.LabelRole)
        
        # Draw the highlight if the track is selected
        if opt.state & QtWidgets.QStyle.State_Selected:
            painter.fillRect(rect, opt.palette.highlight())
            
        # Define padded contents rect
        contents = rect.adjusted(5, 0, -5, 0)
        
        # Draw the track ID
        painter.setPen(QtGui.QPen(opt.palette.color(QtGui.QPalette.Text)))
        painter.drawText(contents, QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter, str(id))
        
        # Draw the track label
        painter.drawText(contents, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter, label)
        
        painter.restore()
    
    def sizeHint(self, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex):
        return QtCore.QSize(100, 30)


from PyQt5 import QtWidgets, QtCore


class FileSelector(QtWidgets.QWidget):
    """Widget to select a file"""
    
    fileSelected = QtCore.pyqtSignal(str)
    
    def __init__(self, parent=None, filter=None):
        super().__init__(parent)
        
        self.filter = filter
        
        self.line_edit = QtWidgets.QLineEdit()
        self.line_edit.setReadOnly(True)
        self.line_edit.setPlaceholderText('Select a file')
        
        self.browse_button = QtWidgets.QPushButton('Browse')
        self.browse_button.clicked.connect(self.browse)
        
        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set horizontal layout
        self.setLayout(QtWidgets.QHBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Arrange line edit and browse button
        self.layout().addWidget(self.line_edit, stretch=1)
        self.layout().addWidget(self.browse_button)
    
    def browse(self):
        """Browse for a file using a file selection dialog"""
        file_name = QtWidgets.QFileDialog.getOpenFileName(self, 'Select a file', filter=self.filter)
        if file_name[0]:
            self.line_edit.setText(file_name[0])
            self.fileSelected.emit(file_name[0])
    


from socket import socket, AF_INET, SOCK_DGRAM, timeout
from typing import Optional
from uuid import uuid4

from sharktopoda_client.model.GenericCommand import GenericCommand
from sharktopoda_client.model.GenericResponse import GenericResponse
from PyQt5 import QtMultimedia, QtCore
from boxjelly.lib.constants import CONTROL_HOST
from boxjelly.lib.settings import SettingsManager

from boxjelly.models.TrackListModel import TrackListModel


class CthulhuVideoController(QtCore.QObject):
    """
    Cthulhu video controller. Redirects calls to Cthulhu UDP commands/responses.
    """
    
    FRAME_UPDATE_INTERVAL = 100
    
    positionChanged = QtCore.pyqtSignal(int)
    frameUpdated = QtCore.pyqtSignal(int)
    
    def __init__(self, parent, track_model: TrackListModel, selection_model: QtCore.QItemSelectionModel):
        super().__init__(parent)
        
        self._track_model = track_model
        self._selection_model = selection_model
        
        self._udp_socket = socket(AF_INET, SOCK_DGRAM)
        self._udp_socket.settimeout(0.5)  # 500ms timeout for response
        
        self._url = None
        self._uuid = None
        
        self._dummy_player = QtMultimedia.QMediaPlayer(self)
        
        self._last_frame_number = None
        
        self._settings = SettingsManager.get_instance()
        
        # Start polling to update the current frame every FRAME_UPDATE_INTERVAL milliseconds
        self.frame_update_timer = QtCore.QTimer(self)
        self.frame_update_timer.timeout.connect(self._do_frame_update)
        self.frame_update_timer.start(CthulhuVideoController.FRAME_UPDATE_INTERVAL)
    
    def send(self, command: GenericCommand) -> None:
        try:
            self._udp_socket.sendto(command.to_json().encode('utf-8'), (CONTROL_HOST, self._settings.control_port.value))
        except OSError as e:
            print(f'Error ({e}) while sending command to Cthulhu:\n{command.to_json(indent=2)}')  # TODO: Log
    
    def send_and_wait(self, command: GenericCommand) -> Optional[GenericResponse]:
        # Send the command
        self.send(command)
        
        # Wait for a response
        try:
            response_raw = self._udp_socket.recv(4096)
            return GenericResponse.from_json(response_raw.decode('utf-8'))
        except timeout:
            # print(f'Timeout while waiting for response to command:\n{command.to_json(indent=2)}')  # TODO: Log
            return None
    
    def set_media(self, media: QtMultimedia.QMediaContent):
        self._url = media.canonicalUrl().toString()
        self._uuid = str(uuid4())
        
        # Update track model video data
        self._track_model.set_video_reference_uuid(self._uuid)
        self._track_model.set_frame_rate(self._settings.frame_rate.value)
        
        self.open(self._url, self._uuid)
        self.pause()
    
    def open(self, url: str, uuid: str):
        """
        Open a video file by URL and assign it a given UUID.
        """
        open_video_command = GenericCommand(command='open', url=url, uuid=uuid)
        self.send_and_wait(open_video_command)
    
    def play(self):
        """
        Play (resume) the current video.
        """
        play_command = GenericCommand(command='play', uuid=self._uuid)
        self.send_and_wait(play_command)
    
    def pause(self):
        """
        Pause the current video.
        """
        pause_command = GenericCommand(command='pause', uuid=self._uuid)
        self.send_and_wait(pause_command)
    
    def stop(self):
        """
        Stop the current video (pause and seek to beginning).
        """
        self.pause()
        
        seek_beginning_command = GenericCommand(command='seek elapsed time', uuid=self._uuid, elapsedTime=0)
        self.send(seek_beginning_command)
    
    def set_frame(self, frame: int):
        """
        Seek the current video to a given frame.
        """
        seek_command = GenericCommand(command='seek elapsed time', uuid=self._uuid, elapsedTime=self._track_model.frame_to_ms(frame))
        self.send(seek_command)
    
    def get_elapsed_time(self):
        """
        Get the current elapsed time in milliseconds of the current video.
        """
        elapsed_time_command = GenericCommand(command='request elapsed time', uuid=self._uuid)
        res = self.send_and_wait(elapsed_time_command)
        
        if not res:
            return
        
        return res.elapsedTime
    
    def get_frame_number(self) -> Optional[int]:
        """
        Get the current frame number of the current video.
        """
        if not (self._url and self._uuid):
            return
        
        elapsed_time = self.get_elapsed_time()
        if not elapsed_time:
            return
        
        return self._track_model.ms_to_frame(elapsed_time)
    
    @QtCore.pyqtSlot()
    def _do_frame_update(self):
        """
        Slot for frame update timer. Updates the current frame.
        """
        frame_number = self.get_frame_number()
        if frame_number is not None and (self._last_frame_number is None or frame_number != self._last_frame_number):
            self.frameUpdated.emit(frame_number)
            self._track_model.load_at_frame(frame_number)
            self._last_frame_number = frame_number




from PyQt5 import QtWidgets, QtCore, QtGui

from boxjelly.ui.settings.tabs.AbstractSettingsTab import AbstractSettingsTab
from boxjelly.ui.settings.tabs.NetworkSettingsTab import NetworkSettingsTab
from boxjelly.ui.settings.tabs.VideoSettingsTab import VideoSettingsTab


class SettingsDialog(QtWidgets.QDialog):
    """
    Settings dialog.

    Contains settings for the application.
    """
    
    applySettings = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        
        self.setWindowTitle('Settings')
        
        # Create tab widget to hold settings pages
        self._tab_widget = QtWidgets.QTabWidget()
        
        # Create button box (OK, Apply, Cancel)
        self._button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok | 
            QtWidgets.QDialogButtonBox.StandardButton.Apply | 
            QtWidgets.QDialogButtonBox.StandardButton.Cancel
        )
        self._button_box.button(QtWidgets.QDialogButtonBox.StandardButton.Ok).pressed.connect(self.accept)
        self._button_box.button(QtWidgets.QDialogButtonBox.StandardButton.Apply).pressed.connect(self._on_apply_pressed)
        self._button_box.button(QtWidgets.QDialogButtonBox.StandardButton.Cancel).pressed.connect(self.reject)
        
        # self.accepted.connect(self.close)
        self.accepted.connect(self._apply)
        self._needs_apply = False
        self._update_apply_enabled()
        
        # Add tabs
        self._add_tabs()
        
        # Arrange the dialog layout
        self._arrange()
        
    def _arrange(self):
        layout = QtWidgets.QVBoxLayout()
        
        layout.addWidget(self._tab_widget)
        layout.addWidget(self._button_box)
        
        self.setLayout(layout)
    
    def _update_apply_enabled(self):
        self._button_box.button(QtWidgets.QDialogButtonBox.StandardButton.Apply).setEnabled(self._needs_apply)
    
    def _on_settings_changed(self):
        self._needs_apply = True
        self._update_apply_enabled()
    
    def _on_apply_pressed(self):
        self._apply()
        
        self._needs_apply = False
        self._update_apply_enabled()
    
    def _apply(self):
        if self._needs_apply:
            self.applySettings.emit()
    
    def _register_tab(self, tab: AbstractSettingsTab):
        # Add tab to tab widget
        if tab.icon is not None:
            self._tab_widget.addTab(tab, tab.icon, tab.name)
        else:
            self._tab_widget.addTab(tab, tab.name)
            
        # Connect signals/slots
        tab.settingsChanged.connect(self._on_settings_changed)
        self.applySettings.connect(tab.apply_settings)
    
    def _add_tabs(self):
        network_tab = NetworkSettingsTab()
        self._register_tab(network_tab)
        
        video_tab = VideoSettingsTab()
        self._register_tab(video_tab)



"""
Track-related models.
"""

from typing import List, Optional
from dataclasses import dataclass

from dataclasses_json import dataclass_json

@dataclass_json
@dataclass
class BoundingBox:
    x: int
    y: int
    w: int
    h: int


@dataclass_json
@dataclass
class Track:
    start_frame: int
    boxes: List[Optional[BoundingBox]]
    
    def __len__(self) -> int:
        return len(self.boxes)
    
    @property
    def slice(self):
        return slice(self.start_frame, self.start_frame + len(self))


@dataclass_json
@dataclass
class IdentifiedTrack(Track):
    id: int
    label: str



import math

from PyQt5 import QtWidgets, QtCore, QtGui


class Timeline(QtWidgets.QWidget):
    
    frameChanged = QtCore.pyqtSignal(int)
    frameStepChanged = QtCore.pyqtSignal(int)
    frameScaleChanged = QtCore.pyqtSignal(float)
    frameSelected = QtCore.pyqtSignal(int)
    
    def __init__(self, parent, frame_step: int, frame_scale: float = 1.0):
        super().__init__(parent)
        
        self._background_brush = self.palette().window()
        self._foreground_brush = self.palette().windowText()
        
        self._frame = 0
        self._frame_step = frame_step
        self._frame_scale = frame_scale
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

    def set_frame(self, frame: int):
        """
        Set the current frame to be displayed.
        """
        if frame != self._frame:
            self.frameChanged.emit(frame)
        self._frame = frame
        self.update()
    
    def set_frame_step(self, frame_step: int):
        """
        Set the frame step.
        """
        if frame_step != self._frame_step:
            self.frameStepChanged.emit(frame_step)
        self._frame_step = frame_step
        self.update()
    
    def set_frame_scale(self, frame_scale: float):
        """
        Set the frame scale.
        """
        if frame_scale != self._frame_scale:
            self.frameScaleChanged.emit(frame_scale)
        self._frame_scale = frame_scale
        self.update()
    
    def map_frame_to_x(self, frame: int) -> int:
        """
        Map a frame to an x coordinate.
        """
        return int((frame - self._frame) * self._frame_scale)
    
    def map_x_to_frame(self, x: int) -> int:
        """
        Map an x coordinate to a frame.
        """
        return int(x / self._frame_scale + self._frame)

    @property
    def frame(self):
        return self._frame
    
    @property
    def frame_step(self):
        return self._frame_step
    
    @property
    def frame_scale(self):
        return self._frame_scale
    
    @property
    def max_frame(self) -> int:
        return int(self._frame + self.width() / self._frame_scale)

    def paintEvent(self, event: QtGui.QPaintEvent):
        # Get the painter
        painter = QtGui.QPainter(self)
        
        # Get the device size
        width = painter.device().width()
        height = painter.device().height()
        
        # Set pen to foreground
        painter.setPen(self._foreground_brush.color())
        
        # Draw horizontal line at bottom
        painter.drawLine(0, height - 1, width, height - 1)
        
        # Compute the multiples of the frame step in the range [frame, max_frame]
        tick_frames = range(math.floor(self._frame / self._frame_step) * self._frame_step, self.max_frame + 1, self._frame_step)
        
        # Draw vertical lines and labels at the tick frames
        for tick_frame in tick_frames:
            x = self.map_frame_to_x(tick_frame)
            painter.drawLine(x, height - 1, x, height - height // 4)
            painter.drawText(x + 5, height - 5, str(tick_frame))
    
    def sizeHint(self):
        return QtCore.QSize(100, 20)
    
    def mousePressEvent(self, event: QtGui.QMouseEvent):
        if event.button() == QtCore.Qt.LeftButton:
            self.frameSelected.emit(self.map_x_to_frame(event.x()))



from PyQt5 import QtWidgets, QtCore, QtGui

from boxjelly.lib.track import IdentifiedTrack
from boxjelly.lib.util import concept_color


class TimelineTrack(QtWidgets.QGraphicsObject):
    """
    Graphical representation of a track within the timeline view.
    """
    
    DEFAULT_HEIGHT = 30
    
    clicked = QtCore.pyqtSignal(object)
    
    def __init__(self, track: IdentifiedTrack, scale_x: float = 1.0, scale_y: float = 1.0, parent=None):
        super().__init__(parent)
        
        self._track = track  # original track
        
        self._scale_x = scale_x  # horizontal scale, in scene units per frame
        self._scale_y = scale_y  # vertical scale, in scene units per height unit
        
        self._track_rect = QtWidgets.QGraphicsRectItem(self)
        self._track_id_text = QtWidgets.QGraphicsSimpleTextItem(str(track.id), self)
        
        self.update()
        
    def update(self):
        self._track_rect.setBrush(concept_color(self._track.label))
        self._rescale()
        super().update()
    
    def set_scale(self, scale_x: float, scale_y: float):
        """
        Set the scale of the track.
        """
        self._scale_x = scale_x
        self._scale_y = scale_y
        
        self._rescale()
    
    def __len__(self):
        return len(self._track)
    
    @property
    def width_scene(self):
        return len(self) * self._scale_x
    
    @property
    def height_scene(self):
        return self.DEFAULT_HEIGHT * self._scale_y 
    
    def _rescale(self):
        """
        Rescale the child items.
        """
        self._track_rect.setRect(0, 0, self.width_scene, self.height_scene)
        
        rect_bounds = self._track_rect.boundingRect()
        text_bounds = self._track_id_text.boundingRect()
        self._track_id_text.setPos(
            int(rect_bounds.width() / 2 - text_bounds.width() / 2), 
            int(rect_bounds.height() / 2 - text_bounds.height() / 2)
        )
        self._track_id_text.setVisible(text_bounds.width() < rect_bounds.width())
            
    @property
    def track(self):
        return self._track
    
    def boundingRect(self) -> QtCore.QRectF:
        return self._track_rect.boundingRect()
    
    def mousePressEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        if event.button() == QtCore.Qt.LeftButton:
            self.clicked.emit(self)
        return super().mousePressEvent(event)
    
    def paint(self, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionGraphicsItem, widget: QtWidgets.QWidget) -> None:
        pass  # don't paint anything. we're just a container for child items.

    


from PyQt5 import QtWidgets
from boxjelly.lib.track import IdentifiedTrack

from boxjelly.models.TrackListModel import TrackListModel


class TrackListModelCommand(QtWidgets.QUndoCommand):
    def __init__(self, track_model: TrackListModel):
        super().__init__()
        
        self._track_model = track_model
        
        self.setText('Track list model command')


class DeleteTrackByRowCommand(TrackListModelCommand):
    def __init__(self, track_model: TrackListModel, row: int):
        super().__init__(track_model)
        
        self._row = row
        self._track = None
        
        self.setText('Delete track by row {}'.format(row))
    
    def undo(self):
        self._track_model.add_track(self._track)
        self._row = self._track_model.index_by_id(self._track.id)
    
    def redo(self):
        self._track = self._track_model.get_track(self._row)
        self._track_model.delete_track(self._row)


class DeleteTrackByIDCommand(TrackListModelCommand):
    def __init__(self, track_model: TrackListModel, track_id: int):
        super().__init__(track_model)
        
        self._track_id = track_id
        self._track = None
        
        self.setText('Delete track {}'.format(track_id))

    def undo(self):
        self._track_model.add_track(self._track)

    def redo(self):
        idx = self._track_model.index_by_id(self._track_id)
        self._track = self._track_model.get_track(idx)
        self._track_model.delete_track(idx)
        

class BatchDeleteTracksByIDCommand(TrackListModelCommand):
    def __init__(self, track_model: TrackListModel, track_ids: list):
        super().__init__(track_model)
        
        self._track_ids = track_ids
        self._tracks = []
        
        self.setText('Batch delete tracks {}'.format(', '.join(str(track_id) for track_id in track_ids)))
    
    def undo(self):
        for track in self._tracks:
            self._track_model.add_track(track)
    
    def redo(self):
        self._tracks.clear()
        for track_id in self._track_ids:
            row_idx = self._track_model.index_by_id(track_id)
            track = self._track_model.get_track(row_idx)
            self._tracks.append(track)
            self._track_model.delete_track(row_idx)
        

class BatchRenameTracksCommand(TrackListModelCommand):
    def __init__(self, track_model: TrackListModel, track_ids: list, new_label: str):
        super().__init__(track_model)
        
        self._track_ids = track_ids
        self._new_label = new_label
        self._old_labels = []
        
        self.setText('Rename track(s) {} -> "{}"'.format(', '.join(str(track_id) for track_id in track_ids), new_label))
    
    def undo(self):
        for track_id, label in zip(self._track_ids, self._old_labels):
            idx = self._track_model.index_by_id(track_id)
            index = self._track_model.index(idx, 0)
            self._track_model.setData(index, label, TrackListModel.LabelRole)
    
    def redo(self):
        self._old_labels.clear()
        for track_id in self._track_ids:
            row_idx = self._track_model.index_by_id(track_id)
            index = self._track_model.index(row_idx, 0)
            self._old_labels.append(self._track_model.data(index, TrackListModel.LabelRole))
            self._track_model.setData(index, self._new_label, TrackListModel.LabelRole)


class SplitTrackCommand(TrackListModelCommand):
    def __init__(self, track_model: TrackListModel, track_id: int, split_idx: int):
        super().__init__(track_model)
        
        self._track_id = track_id
        self._new_track_id = None
        self._split_idx = split_idx
        
        self.setText('Split track {} at {}'.format(track_id, split_idx))
    
    def undo(self):
        old_track_row_idx = self._track_model.index_by_id(self._track_id)
        old_track_index = self._track_model.index(old_track_row_idx, 0)
        
        new_track_row_idx = self._track_model.index_by_id(self._new_track_id)
        new_track_index = self._track_model.index(new_track_row_idx, 0)
        
        old_boxes = old_track_index.data(TrackListModel.BoxesRole)
        new_boxes = new_track_index.data(TrackListModel.BoxesRole)
        
        self._track_model.setData(old_track_index, old_boxes + new_boxes, TrackListModel.BoxesRole)  # Concatenate new boxes back to old track
        self._track_model.delete_track(new_track_row_idx)  # Remove new track
    
    def redo(self):
        row_idx = self._track_model.index_by_id(self._track_id)
        index = self._track_model.index(row_idx, 0)
        
        boxes = self._track_model.data(index, TrackListModel.BoxesRole)
        
        old_boxes = boxes[:self._split_idx]
        new_boxes = boxes[self._split_idx:]
        
        old_start_frame = index.data(TrackListModel.StartFrameRole)
        old_label = index.data(TrackListModel.LabelRole)
        
        new_track = IdentifiedTrack(
            old_start_frame + self._split_idx,  # Start frame = old start frame + split index
            new_boxes,
            self._track_model.get_next_id(),  # ID = next ID available from model
            old_label  # Label = old label
        )
        
        self._new_track_id = new_track.id
        
        self._track_model.setData(index, old_boxes, TrackListModel.BoxesRole)  # Clip the old track boxes
        self._track_model.add_track(new_track)  # Add the new track


class MergeTracksCommand(TrackListModelCommand):
    def __init__(self, track_model: TrackListModel, track_ids: list):
        super().__init__(track_model)
        
        self._track_ids = track_ids
        self._original_tracks = []
        self._new_track_id = None
        
        self.setText('Merge tracks {}'.format(', '.join(str(track_id) for track_id in track_ids)))
    
    def undo(self):
        new_track_row_idx = self._track_model.index_by_id(self._new_track_id)
        self._track_model.delete_track(new_track_row_idx)  # Delete the new track
        
        # Restore the original tracks
        for track in self._original_tracks:
            self._track_model.add_track(track)
    
    def redo(self):
        # Collect the tracks and remove them from the model
        self._original_tracks.clear()
        for track_id in self._track_ids:
            row_idx = self._track_model.index_by_id(track_id)
            track = self._track_model.get_track(row_idx)
            self._original_tracks.append(track)
            self._track_model.delete_track(row_idx)
            
        # Sort by start frame
        self._original_tracks.sort(key=lambda track: track.start_frame)
        
        start_frame = self._original_tracks[0].start_frame
        end_frame = max(track.start_frame + len(track.boxes) for track in self._original_tracks)
        
        boxes_by_frame = []
        next_track_idx = 0
        for frame_num in range(start_frame, end_frame + 1):
            if next_track_idx < len(self._original_tracks) and self._original_tracks[next_track_idx].start_frame == frame_num:
                current_track = self._original_tracks[next_track_idx]
                next_track_idx += 1
            
            # Get the index of the box in the current track that corresponds to this frame
            box_idx = frame_num - current_track.start_frame
            
            # Append the corresponding box (or None if there is no box) to the list
            if box_idx < len(current_track.boxes):
                boxes_by_frame.append(current_track.boxes[box_idx])
            else:  # If the box index is out of bounds, then the current track has no boxes for this frame
                boxes_by_frame.append(None)
        
        # Create the new track
        new_track = IdentifiedTrack(
            start_frame,  # Start frame = start frame of first track
            boxes_by_frame,
            self._track_model.get_next_id(),  # ID = next ID available from model
            self._original_tracks[0].label  # Label = label of first track
        )
        
        self._new_track_id = new_track.id

        self._track_model.add_track(new_track)  # Add the new track
        


import logging
from typing import Iterable, List, Set
from uuid import UUID, uuid4
from intervaltree.interval import Interval
from intervaltree.intervaltree import IntervalTree

from PyQt5 import QtCore
from sharktopoda_client.localization.LocalizationController import LocalizationController
from sharktopoda_client.localization.Localization import Localization
from sharktopoda_client.localization.IO import IO as LocalizationIO
from boxjelly.lib.settings import SettingsManager

from boxjelly.lib.track import IdentifiedTrack


class TrackListModel(QtCore.QAbstractListModel):
    """
    Track list model. Wraps a list of identified tracks and synchronizes with a Sharktopoda client LocalizationController.
    """
    
    IDRole = QtCore.Qt.UserRole + 1  # Integer ID (e.g. 42)
    LabelRole = QtCore.Qt.UserRole + 2  # String label (e.g. "fish")
    StartFrameRole = QtCore.Qt.UserRole + 3  # Integer start frame (e.g. 12345)
    EndFrameRole = QtCore.Qt.UserRole + 4  # Integer end frame (e.g. 67890). This is derived, so cannot be set directly.
    BoxesRole = QtCore.Qt.UserRole + 5  # List of track's boxes (None indicates no box at that frame)
    
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        
        self._tracks: List[IdentifiedTrack] = []  # Root list of IdentifiedTracks
        
        # IntervalTree for quick lookup of tracks by frame
        self._track_id_interval_dict = {}
        self._track_tree = IntervalTree()
        
        # Sharktopoda client LocalizationController for adding/removing localizations
        self._track_id_localization_list_dict = {}
        self._localization_controller = LocalizationController()
        self._localization_controller.log.setLevel(logging.INFO)
        
        # Get the app settings manager, connect signals from setting proxies
        self._settings = SettingsManager.get_instance()
        self._settings.incoming_port.valueChanged.connect(self._refresh_localization_io)
        self._settings.incoming_topic.valueChanged.connect(self._refresh_localization_io)
        self._settings.outgoing_port.valueChanged.connect(self._refresh_localization_io)
        self._settings.outgoing_topic.valueChanged.connect(self._refresh_localization_io)
        
        # Create the localization IO
        self._localization_io = None
        self._selection_controller = None
        self._refresh_localization_io()
        
        # Video info (TODO: find this a new home)
        self._video_reference_uuid = None
        self._frame_rate = None
        self._settings.frame_rate.valueChanged.connect(self.set_frame_rate)
        
        # Set of track IDs that have been loaded into the LocalizationController
        self._loaded_ids = set()  # Set of track IDs that are currently loaded in Cthulhu
    
    @QtCore.pyqtSlot()
    def _refresh_localization_io(self):
        """
        Refresh the LocalizationIO.
        """
        # Close the current IO
        old_io = self._localization_io
        if old_io is not None:
            old_io.ok = False
            old_io.context.destroy()
        
        # Create a new IO from the current settings
        new_io = LocalizationIO(
            self._settings.incoming_port.value,
            self._settings.outgoing_port.value,
            self._settings.incoming_topic.value,
            self._settings.outgoing_topic.value,
            self._localization_controller
        )
        
        # Replace the old IO with the new one
        self._localization_io = new_io
        self._localization_io.log.setLevel(logging.INFO)
        self._selection_controller = self._localization_io.selectionController
    
    @property
    def video_loaded(self) -> bool:
        """
        Property to tell whether or not the video is loaded.
        """
        return self._video_reference_uuid is not None and self._frame_rate is not None
    
    def set_video_reference_uuid(self, uuid: str):
        """
        Set the video reference UUID.
        """
        self._video_reference_uuid = uuid
    
    def set_frame_rate(self, frame_rate: float):
        """
        Set the video frame rate.
        """
        # Check if we need to reload tracks
        if self._frame_rate is not None and self._frame_rate != frame_rate:
            self._localization_controller.clear()
            self._loaded_ids.clear()
        
        self._frame_rate = frame_rate
    
    def frame_to_ms(self, frame: int):
        """
        Convert a frame number to elapsed milliseconds.
        """
        if not self.video_loaded:
            return ValueError('Video not loaded')
        
        return round((1e3 / self._frame_rate) * frame)

    def ms_to_frame(self, ms: int):
        """
        Convert elapsed milliseconds to a frame number.
        """
        if not self.video_loaded:
            return ValueError('Video not loaded')
        
        return round(self._frame_rate * (ms / 1e3))
    
    def sort(self, *_):
        """
        Sort the list of tracks by start frame.
        """
        self._tracks.sort(key=lambda track: (track.start_frame, track.id))
        self.layoutChanged.emit()

    def rowCount(self, parent) -> int:
        """
        Get the number of tracks in the list.
        """
        return len(self._tracks)

    def _register(self, track: IdentifiedTrack):
        """
        Register a track internally. Do not call this directly - use add_track or set_tracks instead.
        """
        # Add the track to the list
        self._tracks.append(track)
        
        # Create an interval that represents the track duration
        interval = Interval(track.start_frame, track.start_frame + len(track), track)
        self._track_tree.add(interval)  # Add the interval to the IntervalTree
        self._track_id_interval_dict[track.id] = interval  # Keep a reference to the interval
        
        # Generate a list of Localizations for the track, but don't add them to the LocalizationController yet
        localizations = list(self._generate_localizations(track))
        self._track_id_localization_list_dict[track.id] = localizations
        
    def _unregister(self, idx: int):
        """
        Unregister a track internally. Do not call this directly - use remove_track instead.
        """
        track_id = self.get_track(idx).id
        
        # Immediately remove the track from the LocalizationController, if loaded
        if track_id in self._loaded_ids:
            self._localization_controller.removeLocalizations(self._track_id_localization_list_dict[track_id])
            self._loaded_ids.remove(track_id)
        del self._track_id_localization_list_dict[track_id]
        
        # Remove the track's interval from the IntervalTree
        interval = self._track_id_interval_dict[track_id]
        self._track_tree.remove(interval)
        del self._track_id_interval_dict[track_id]
        
        # Remove the track from the list
        del self._tracks[idx]

    def add_track(self, track: IdentifiedTrack):
        """
        Add a track to the list.
        
        Note: this will emit the layoutChanged signal on each call, which is inefficient when rendering to multiple displays. 
        If you want to add multiple tracks at once, use set_tracks instead.
        """
        self.beginInsertRows(QtCore.QModelIndex(), len(self._tracks), len(self._tracks))
        self._register(track)
        self.endInsertRows()
        self.sort()
    
    def clear(self):
        """
        Clear the list of tracks.
        """
        self.beginResetModel()
        self._tracks = []
        self._track_tree.clear()
        self._track_id_interval_dict = {}
        self._track_id_localization_list_dict = {}
        self._loaded_ids = set()
        self._localization_controller.clear()
        self.endResetModel()
    
    def set_tracks(self, tracks: List[IdentifiedTrack]):
        """
        Set the list of tracks.
        """
        self.clear()
        self.beginInsertRows(QtCore.QModelIndex(), 0, len(tracks) - 1)
        for track in tracks:
            self._register(track)
        self.endInsertRows()
        self.sort()
    
    def _load_localizations(self, track: IdentifiedTrack):
        """
        Load the track's localizations into the LocalizationController.
        """
        if track.id in self._loaded_ids:  # Short-circuit if the track is already loaded
            return
        
        localizations = self._track_id_localization_list_dict[track.id]
        self._localization_controller.addLocalizations(localizations)
        self._loaded_ids.add(track.id)
    
    def load_at_frame(self, frame_number: int):
        """
        Load localizations for tracks overlapping the given frame number.
        """
        overlapping_intervals = self._track_tree[frame_number-100:frame_number+100]
        for interval in overlapping_intervals:
            self._load_localizations(interval.data)
    
    def get_track(self, idx: int) -> IdentifiedTrack:
        """
        Get the track at the given index.
        """
        return self._tracks[idx]
            
    def index_by_id(self, id: int) -> int:
        """
        Find the index of the track with the given ID.
        
        Raises a ValueError if a track with the given ID is not found.
        """
        return self._tracks.index(self._track_id_interval_dict[id].data)
    
    def get_interval(self, key):
        """
        Index the IntervalTree with the given key.
        
        TODO: Don't interfaace with the IntervalTree directly.
        """
        return self._track_tree[key]
    
    def set_track(self, idx: int, track: IdentifiedTrack):
        """
        Update the track at the given index.
        """
        self._tracks[idx] = track
        self.dataChanged.emit(self.index(idx), self.index(idx))
        self.sort()
        
    def delete_track(self, idx):
        """
        Delete the track at the given index.
        """
        self.beginRemoveRows(QtCore.QModelIndex(), idx, idx)
        self._unregister(idx)
        self.endRemoveRows()
        
    def get_all_tracks(self):
        """
        Get the list of all tracks.
        """
        return self._tracks

    def get_next_id(self):
        """
        Get the next available track ID.
        """
        max_id = 0 if len(self._tracks) == 0 else max(track.id for track in self._tracks)
        return max_id + 1

    def data(self, index, role):
        if not index.isValid():
            return None
        
        track = self.get_track(index.row())
        
        if role == QtCore.Qt.DisplayRole:
            return track.label
        elif role == self.IDRole:
            return track.id
        elif role == self.LabelRole:
            return track.label
        elif role == self.StartFrameRole:
            return track.start_frame
        elif role == self.EndFrameRole:
            return track.start_frame + len(track) - 1
        elif role == self.BoxesRole:
            return track.boxes
        return None
    
    def setData(self, index, value, role):
        if not index.isValid():
            return False
        
        track = self.get_track(index.row())
        
        if role == self.LabelRole:
            # Update the label
            track.label = value
            
            # Replace localizations
            localizations = self._track_id_localization_list_dict[track.id]  # Get old localizations
            new_localizations = list(self._generate_localizations(track))  # Make the new localizations
            self._localization_controller.removeLocalizations(localizations)  # Remove the old localizations
            self._localization_controller.addLocalizations(new_localizations)  # Add the new localizations
            self._track_id_localization_list_dict[track.id] = new_localizations  # Update the localization list dict
            
            self.dataChanged.emit(index, index)
        elif role == self.BoxesRole:    
            # Update the boxes
            track.boxes = value
            
            # Replace interval
            interval = self._track_id_interval_dict[track.id]  # Get old interval
            new_interval = Interval(interval.begin, interval.begin + len(value), track)  # Make the new interval
            self._track_tree.remove(interval)  # Remove the old interval
            self._track_tree.add(new_interval)  # Add the new interval
            self._track_id_interval_dict[track.id] = new_interval  # Update the interval dict
            
            # Replace localizations
            localizations = self._track_id_localization_list_dict[track.id]  # Get old localizations
            new_localizations = list(self._generate_localizations(track))  # Make the new localizations
            self._localization_controller.removeLocalizations(localizations)  # Remove the old localizations
            self._localization_controller.addLocalizations(new_localizations)  # Add the new localizations
            self._track_id_localization_list_dict[track.id] = new_localizations  # Update the localization list dict
            
            self.dataChanged.emit(index, index)
        else:
            return False
        
        return True

    def _generate_localizations(self, track: IdentifiedTrack) -> Iterable[Localization]:
        """
        Generate localizations for the given track.
        """
        for frame_offset, box in enumerate(track.boxes):
            if not box:
                continue
            yield Localization(
                concept=f'{track.id} - {track.label}',
                elapsedTime=self.frame_to_ms(track.start_frame + frame_offset),
                duration=0,
                videoReferenceUuid=UUID(self._video_reference_uuid),
                annotationUuid=UUID(int=track.id),
                localizationUuid=uuid4(),  # TODO: Ensure unique
                x=box.x,
                y=box.y,
                width=box.w,
                height=box.h
            )
    
    def select_track(self, idx: int):
        """
        Select the track at the given index.
        """
        selected_track = self.get_track(idx)
        if selected_track.id not in self._loaded_ids:  # track needs load
            self._load_localizations(selected_track)
        
        track_localizations = self._track_id_localization_list_dict[selected_track.id]
        
        # Sharktopoda client clears the selection. We avoid this by unioning the new selection with the old selection.
        previous_localizations = self._selection_controller.getSelectedLocalizations()
        previous_localization_uuids = set(localization.localizationUuid for localization in previous_localizations)
        
        union = previous_localizations + [loc for loc in track_localizations if loc.localizationUuid not in previous_localization_uuids]
        
        self._selection_controller.select(union, True)
    
    def deselect_track(self, idx: int):
        """
        Deselect the track at the given index.
        """
        deselected_track = self.get_track(idx)
        if deselected_track.id not in self._loaded_ids:  # track needs load
            self._load_localizations(deselected_track)
        
        track_localizations = self._track_id_localization_list_dict[deselected_track.id]
        
        self._selection_controller.deselect(track_localizations, True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boxjelly',
 'boxjelly.commands',
 'boxjelly.delegates',
 'boxjelly.lib',
 'boxjelly.models',
 'boxjelly.scripts',
 'boxjelly.ui',
 'boxjelly.ui.graphicsitems',
 'boxjelly.ui.settings',
 'boxjelly.ui.settings.tabs',
 'boxjelly.ui.track',
 'boxjelly.ui.video']

package_data = \
{'': ['*'], 'boxjelly': ['assets/*', 'assets/icons/*', 'assets/images/*']}

install_requires = \
['dataclasses-json>=0.5.7,<0.6.0',
 'intervaltree>=3.1.0,<4.0.0',
 'pyqt5>=5.15,<6.0',
 'sharktopoda-client>=0.1.4,<0.2.0']

entry_points = \
{'console_scripts': ['boxjelly = boxjelly.scripts.run:main']}

setup_kwargs = {
    'name': 'boxjelly',
    'version': '0.1.4',
    'description': 'BoxJelly is a tool for viewing and editing object tracks in video.',
    'long_description': '![BoxJelly logo](boxjelly/assets/images/boxjelly_logo_128.png)\n\n# BoxJelly\n\n**BoxJelly** is a tool for viewing and editing object tracks in video.\n\n[![MIT License](https://img.shields.io/badge/license-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n[![Python](https://img.shields.io/badge/language-Python-blue.svg)](https://www.python.org/downloads/)\n\nAuthor: Kevin Barnard, [kbarnard@mbari.org](mailto:kbarnard@mbari.org)\n\n---\n\n## Cthulhu configuration **(required)**\n\nRecently, BoxJelly ditched its internal video player in favor of [Cthulhu](https://github.com/mbari-media-management/cthulhu). This integration is still in development and has some limitations.\n\nAs a result, you must configure Cthulhu and BoxJelly before they can be used together. The following configuration is required:\n\n1. **Set the BoxJelly framerate**. Cthulhu does not report video framerate, so a default of 29.97 is assumed. This is configurable in the BoxJelly settings (Ctrl+,).\n2. **Set the Cthulhu global duration**. Set the appropriate duration for localizations in the Cthulhu "Annotations" settings. Normally, this is `1000/fps`. If you notice flickering, you may need to increase this value. If you notice overlapping boxes within the same track, you may need to decrease this value.\n\n## Install\n\n### From PyPI\n\nBoxJelly is available on PyPI as `boxjelly`. To install, run:\n\n```bash\npip install boxjelly\n```\n\n### From source\n\nThis project is build with Poetry. To install from source, run (in the project root):\n\n```bash\npoetry install\n```\n\n## Run\n\nOnce BoxJelly is installed, you can run it from the command line:\n\n```bash\nboxjelly\n```\n\n**You must have Cthulhu installed and running before you can use BoxJelly.**\n\nDetailed usage is documented in [USAGE](USAGE.md).\n\n---\n\nCopyright &copy; 2021&ndash;2022 [Monterey Bay Aquarium Research Institute](https://www.mbari.org)\n',
    'author': 'Kevin Barnard',
    'author_email': 'kbarnard@mbari.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
File IO operations. 

Each type is encapsulated in a QObject subclass which provides read/write capability.
"""

from collections import defaultdict
from io import BytesIO
import json
import os
import tarfile
from typing import List, Union, Any
from pathlib import Path
from uuid import UUID

from PyQt5 import QtCore

from boxjelly.lib.track import BoundingBox, IdentifiedTrack


class AbstractFileIO(QtCore.QObject):
    """
    File IO abstract class.
    """
    
    fileRead = QtCore.pyqtSignal(object)  # payload can be any object
    fileWritten = QtCore.pyqtSignal()
    fileReadFailure = QtCore.pyqtSignal(str)
    fileWriteFailure = QtCore.pyqtSignal(str)
    
    def __init__(self, path: Union[str, Path]):
        super().__init__()
        
        self._path = None
        self.path = path
    
    @property
    def path(self) -> Path:
        return self._path
    
    @path.setter
    def path(self, path: Union[str, Path]):
        self._path = Path(path)  # ensure type Path
    
    def _write(self, data: Any):
        """
        Write data to file. Do not call this directly; instead call write.
        
        Subclasses should implement this method.
        """
        raise NotImplementedError()
        
    def write(self, data: Any):
        """
        Write data to file.
        """
        try:
            self._write(data)
        except Exception as e:
            self.fileWriteFailure.emit(str(e))
            return
        
        self.fileWritten.emit()
    
    def _read(self) -> Any:
        """
        Read data from file. Do not call this directly; instead call read.
        
        Subclasses should implement this method.
        """
        raise NotImplementedError()
    
    def read(self) -> Any:
        """
        Read data from file.
        """
        try:
            data = self._read()
        except Exception as e:
            self.fileReadFailure.emit(str(e))
            return
        
        self.fileRead.emit(data)
        return data


class AbstractTrackFileIO(AbstractFileIO):
    """
    File IO abstract class that deals with collections of identified tracks.
    """
    
    def _write(self, data: List[IdentifiedTrack]):
        return super()._write(data)
    
    def write(self, data: List[IdentifiedTrack]):
        """
        Write tracks to file.
        """
        return super().write(data)
    
    def _read(self) -> List[IdentifiedTrack]:
        return super()._read()
    
    def read(self) -> List[IdentifiedTrack]:
        """
        Read tracks from file.
        """
        return super().read()


class JSONTrackFileIO(AbstractTrackFileIO):
    """
    JSON track file IO.
    """
    
    def _write(self, data: List[IdentifiedTrack]):
        track_dicts = [track.to_dict() for track in data]
        
        with self._path.open('w') as f:
            json.dump(track_dicts, f, indent=2)
    
    def _read(self) -> List[IdentifiedTrack]:
        with self._path.open('r') as f:
            track_dicts = json.load(f)
        
        tracks = []
        for track_dict in track_dicts:
            tracks.append(IdentifiedTrack.from_dict(track_dict))
        return tracks


class YOLOv5DeepSortTrackFileIO(AbstractTrackFileIO):
    """
    YOLOv5-DeepSort track file IO.
    """
    
    def _write(self, data: List[IdentifiedTrack]):
        raise NotImplementedError()
    
    def _read(self) -> List[IdentifiedTrack]:
        track_dict = {}
        
        with open(self._path, 'r') as f:
            for line in f:
                if not line:  # Empty line, skip
                    continue
                
                # Try to parse line fields
                try:
                    parts = line.strip().split(' ')
                    
                    # video_filename = parts[0]
                    frame_number = int(parts[1])
                    track_id = int(parts[2])
                    x = int(parts[3])
                    y = int(parts[4])
                    w = int(parts[5])
                    h = int(parts[6])
                    label = ' '.join(parts[11:])
                    
                    # Create track if it doesn't exist
                    if track_id not in track_dict:
                        track_dict[track_id] = IdentifiedTrack(start_frame=frame_number, boxes=[], id=track_id, label=label)
                    track = track_dict[track_id]
                    
                    # Fill in frame jump
                    if track.boxes:
                        start_frame = track.start_frame
                        while frame_number != start_frame + len(track.boxes):
                            track.boxes.append(None)
                    
                    # Add detection to track
                    track.boxes.append(BoundingBox(x, y, w, h))
                    
                except Exception as e:
                    raise ValueError(f'Failed to parse YOLOv5-DeepSort line: {line}') from e
        
        tracks = list(track_dict.values())
        
        return tracks


class DeepseaTrackFileIO(AbstractTrackFileIO):
    """
    Deepsea track file IO.
    """
    
    def _write(self, data: List[IdentifiedTrack]):
        # Reformat list of tracks as dict of frame -> "visual event" set
        visual_events_by_frame = defaultdict(list)
        for track in data:
            for frame_offset, box in enumerate(track.boxes):
                frame_num = track.start_frame + frame_offset
                visual_events_by_frame[frame_num].append({
                    'bounding_box': {
                        'height': int(box.h),
                        'width': int(box.w),
                        'x': int(box.x),
                        'y': int(box.y)
                    },
                    'class_index': -1,  # TODO: fill in
                    'class_name': track.label,
                    'confidence': -1,  # TODO: fill in
                    'frame_num': frame_num,
                    'occlusion': -1,  # TODO: fill in
                    'surprise': -1,  # TODO: fill in
                    'uuid': str(UUID(int=track.id))  # TODO: come up with mapping system ... or switch to UUIDs internally
                })
        
        # Get max frame number
        max_frame = max(visual_events_by_frame.keys())
        
        # For each frame, write boxes to JSON in .tar.gz archive
        with tarfile.open(self._path, 'w:gz') as tar:
            for frame in range(1, max_frame + 1):
                visual_events = visual_events_by_frame.get(frame, [])
                filename = f'f{str(frame).zfill(6)}.json'
                
                json_data = [  # wacky format
                    'visualevents',
                    [
                        ['visualevent'] + [visual_event]
                        for visual_event in visual_events
                    ]
                ]
                
                # Encode and describe file
                byte_data = json.dumps(json_data).encode('utf-8')
                buf = BytesIO(byte_data)
                tar_info = tarfile.TarInfo(filename)
                tar_info.size = len(byte_data)
                
                # Add to archive
                tar.addfile(tar_info, buf)
    
    def _read(self) -> List[IdentifiedTrack]:
        track_dict = {}
    
        # Open the tar file
        with tarfile.open(self._path) as t:
            # Find archived JSON files
            json_paths = [path for path in t.getnames() if path.endswith('.json') and os.path.basename(path).startswith('f')]
        
            for json_path in json_paths:
                # Try to parse the JSON
                try:
                    with t.extractfile(json_path) as f:
                        data = json.load(f)
                        event_items = data[1]
                        for event_item in event_items:
                            event_data = event_item[1]
                            
                            frame_number = int(event_data['frame_num'])
                            box = event_data['bounding_box']
                            x = int(box['x'])
                            y = int(box['y'])
                            w = int(box['width'])
                            h = int(box['height'])
                            label = event_data['class_name']
                            
                            # Create track if it doesn't exist
                            uuid = event_data['uuid']
                            if uuid not in track_dict:
                                track_dict[uuid] = IdentifiedTrack(start_frame=frame_number, boxes=[], id=len(track_dict), label=label)
                            track = track_dict[uuid]
                            
                            # Fill in frame jump
                            if track.boxes:
                                start_frame = track.start_frame
                                while frame_number != start_frame + len(track.boxes):
                                    track.boxes.append(None)
                            
                            # Add detection to track
                            track.boxes.append(BoundingBox(x, y, w, h))
                            
                except Exception as e:
                    raise ValueError(f'Failed to parse {self.filename}/{json_path}') from e

        tracks = list(track_dict.values())
        
        return tracks



import hashlib

from PyQt5.QtGui import QColor


def byte_hash(s: str) -> int:
    """
    Compute a one-byte hash of a string.
    """
    return hashlib.sha256(s.encode()).digest()[0]


def concept_color(concept: str) -> QColor:
    """
    Compute a color for a concept based on its name.
    """
    color = QColor()
    color.setHsv(byte_hash(concept) % 256, 255, 255)
    return color.lighter()
